Package: phsopendata
Title: Extract from the Scottish Health and Social Care Open Data
        Platform
Version: 1.1.0
Authors@R: c(
    person("Public Health Scotland", , , "phs.datascience@phs.scot", role = "cph"),
    person("Csilla", "Scharle", , "csilla.scharle2@phs.scot", role = c("cre", "aut")),
    person("James", "Hayes", , "James.Hayes2@phs.scot", role = "aut",
           comment = c(ORCID = "0000-0002-5380-2029")),
    person("David", "Aikman", , "david.aikman@phs.scot", role = "aut"),
    person("Ross", "Hull", , "ross.hull2@phs.scot", role = "aut"),
    person("Haritha", "Jagadeesh", , "haritha.jagadeesh@phs.scot", role = "aut"),
    person("Simon", "Barnes", , "simon.barnes@phs.scot", role = "ctb")
  )
Description: Extract and interact with data from the Scottish Health and
    Social Care Open Data platform <https://www.opendata.nhs.scot>.
License: MIT + file LICENSE
URL: https://github.com/Public-Health-Scotland/phsopendata,
        https://public-health-scotland.github.io/phsopendata/
BugReports: https://github.com/Public-Health-Scotland/phsopendata/issues
Imports: cli (>= 3.2.0), dplyr (>= 1.0.0), httr (>= 1.0.0), lifecycle,
        magrittr (>= 1.0.0), purrr (>= 1.0.0), readr (>= 1.0.0), rlang
        (>= 1.0.0), stringdist, tibble (>= 3.0.0)
Suggests: covr, curl, jsonlite (>= 1.1), spelling, testthat (>= 3.0.0),
        xml2
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-GB
NeedsCompilation: no
Packaged: 2026-05-05 16:02:50 UTC; jamesm32
Author: Public Health Scotland [cph],
  Csilla Scharle [cre, aut],
  James Hayes [aut] (ORCID: <https://orcid.org/0000-0002-5380-2029>),
  David Aikman [aut],
  Ross Hull [aut],
  Haritha Jagadeesh [aut],
  Simon Barnes [ctb]
Maintainer: Csilla Scharle <csilla.scharle2@phs.scot>
Repository: CRAN
Date/Publication: 2026-05-06 08:50:02 UTC
Built: R 4.5.2; ; 2026-05-06 11:34:08 UTC; unix

Package: phsopendata
Title: Extract from the Scottish Health and Social Care Open Data
        Platform
Version: 1.1.0
Authors@R: c(
    person("Public Health Scotland", , , "phs.datascience@phs.scot", role = "cph"),
    person("Csilla", "Scharle", , "csilla.scharle2@phs.scot", role = c("cre", "aut")),
    person("James", "Hayes", , "James.Hayes2@phs.scot", role = "aut",
           comment = c(ORCID = "0000-0002-5380-2029")),
    person("David", "Aikman", , "david.aikman@phs.scot", role = "aut"),
    person("Ross", "Hull", , "ross.hull2@phs.scot", role = "aut"),
    person("Haritha", "Jagadeesh", , "haritha.jagadeesh@phs.scot", role = "aut"),
    person("Simon", "Barnes", , "simon.barnes@phs.scot", role = "ctb")
  )
Description: Extract and interact with data from the Scottish Health and
    Social Care Open Data platform <https://www.opendata.nhs.scot>.
License: MIT + file LICENSE
URL: https://github.com/Public-Health-Scotland/phsopendata,
        https://public-health-scotland.github.io/phsopendata/
BugReports: https://github.com/Public-Health-Scotland/phsopendata/issues
Imports: cli (>= 3.2.0), dplyr (>= 1.0.0), httr (>= 1.0.0), lifecycle,
        magrittr (>= 1.0.0), purrr (>= 1.0.0), readr (>= 1.0.0), rlang
        (>= 1.0.0), stringdist, tibble (>= 3.0.0)
Suggests: covr, curl, jsonlite (>= 1.1), spelling, testthat (>= 3.0.0),
        xml2
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-GB
NeedsCompilation: no
Packaged: 2026-05-05 16:02:50 UTC; jamesm32
Author: Public Health Scotland [cph],
  Csilla Scharle [cre, aut],
  James Hayes [aut] (ORCID: <https://orcid.org/0000-0002-5380-2029>),
  David Aikman [aut],
  Ross Hull [aut],
  Haritha Jagadeesh [aut],
  Simon Barnes [ctb]
Maintainer: Csilla Scharle <csilla.scharle2@phs.scot>
Repository: CRAN
Date/Publication: 2026-05-06 08:50:02 UTC
Built: R 4.6.0; ; 2026-05-06 11:58:48 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(get_dataset)
export(get_dataset_additional_info)
export(get_latest_resource)
export(get_resource)
export(get_resource_sql)
export(list_datasets)
export(list_resources)
importFrom(magrittr,"%>%")

<!-- README.md is generated from README.Rmd. Please edit that file -->

# phsopendata

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/phsopendata)](https://CRAN.R-project.org/package=phsopendata)
[![GitHub release (latest by
date)](https://img.shields.io/github/v/release/Public-Health-Scotland/phsopendata)](https://github.com/Public-Health-Scotland/phsopendata/releases/latest)
[![R-CMD-check](https://github.com/Public-Health-Scotland/phsopendata/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/Public-Health-Scotland/phsopendata/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/Public-Health-Scotland/phsopendata/branch/master/graph/badge.svg)](https://app.codecov.io/gh/Public-Health-Scotland/phsopendata?branch=master)

<!-- badges: end -->

`phsopendata` contains functions to interact with open data from the
[Scottish Health and Social Care Open Data
platform](https://www.opendata.nhs.scot/) via the CKAN API.

- `get_resource()` extracts a single resource from an open dataset by
  resource id
- `get_latest_resource()` extracts the most recent resource from
  applicable datasets, by dataset name
- `get_dataset()` extracts multiple resources from an open dataset by
  dataset name
- `list_datasets()` returns the names of all available datasets
- `list_resources()` returns information on all resources within an open
  dataset by dataset name

`phsopendata` can be used on both Posit Workbench and desktop versions
of RStudio.

## Installation

``` r
# The easiest way to get phsopendata is to install from CRAN:
install.packages("phsopendata")
```

### Development version

To get a bug fix or to use a feature from the development version, you
can install the development version of phsopendata from GitHub.

``` r
# install.packages("remotes")
remotes::install_github("Public-Health-Scotland/phsopendata")
```

## Examples

### Downloading a data table with `get_resource()`

To extract a specific resource, you will need its unique identifier -
resource id. This can be found in the dataset metadata, the URL of a
resource’s page on <https://www.opendata.nhs.scot/>, or extracted using
`list_resources()`.

``` r
library(phsopendata)

# define a resource ID
res_id <- "a794d603-95ab-4309-8c92-b48970478c14"

# download the data from the CKAN database
data <- get_resource(res_id = "a794d603-95ab-4309-8c92-b48970478c14")
```

### Querying/filtering data with `get_resource()`

You can define a row limit with the `rows` argument to get the first *N*
rows of a table.

``` r
# get first 100 rows
get_resource(
  res_id = "a794d603-95ab-4309-8c92-b48970478c14",
  rows = 100
)
```

You can use `col_select` and `row_filters` to query the data server-side
(i.e., the data is filtered before it is downloaded to your machine).

``` r
# get first 100 rows
get_resource(
  res_id = "a794d603-95ab-4309-8c92-b48970478c14",
  col_select = c("GPPracticeName", "TelephoneNumber"),
  row_filters = list(
    HB = "S08000017",
    Dispensing = "Y"
  )
)
```

### Downloading multiple tables with `get_dataset()`

To extract all resources from a dataset, you will need to use the
*dataset name*. Note that this will differ from the *dataset title* that
displays on the website. This can be found using `list_datasets()`, or
taken from the dataset URL.

In this example, we are downloading GP Practice Population Demographics
from:
[opendata.nhs.scot/dataset/*gp-practice-populations*](https://www.opendata.nhs.scot/dataset/gp-practice-populations),
so the dataset name will be gp-practice-populations.

``` r
# if max_resources is not set, all resources will be returned by default.
# Here we pull 10 rows from the first 2 resources only
get_dataset("gp-practice-populations", max_resources = 2, rows = 10)
```

## Contributing to phsopendata

At present, this package is maintained by [James
Hayes](https://github.com/Moohan).

If you have requests or suggestions for additional functionality, please
contact the package maintainer and/or the [PHS Open Data
team](mailto:phs.opendata@phs.scot).

If you would like to share examples of how you work with open data, you
can also do so in the [Open Data
repository](https://github.com/Public-Health-Scotland/Open-Data), where
example scripts and resources are collated.

#' Create fields query for GET
#'
#' @description Produces a comma separated list of column names for use in a GET request.
#'
#' @param col_select A character vector containing the names of desired columns/fields (e.g., c("Date", "Sex")).
#'
#' @return a character string
#' @noRd
#' @keywords internal
parse_col_select <- function(col_select, call = rlang::caller_env()) {
  if (is.null(col_select)) {
    return(NULL)
  }

  if (inherits(col_select, "list")) {
    col_select <- unlist(col_select)
  }

  # Remove any duplicates
  col_select <- unique(col_select)

  if (!inherits(col_select, "character")) {
    cli::cli_abort(
      "{.arg col_select} must be a {.cls character} vector, not a {.cls {class(col_select)}} vector.",
      call = call
    )
  }

  return(paste(col_select, collapse = ","))
}

test_that("creates commas separated string from vector", {
  expect_identical(
    parse_col_select(letters[1:5]),
    "a,b,c,d,e"
  )
})

test_that("returns NULL if input is NULL", {
  expect_null(parse_col_select(NULL))
})

test_that("Can deal with a list in some cases", {
  expect_identical(
    parse_col_select(list("col1", "col2")),
    "col1,col2"
  )
  expect_identical(
    parse_col_select(list(columns = c("col1", "col2"))),
    "col1,col2"
  )
  expect_identical(
    parse_col_select(list(
      columns = c("col1", "col2"),
      columns2 = c("col1", "col2")
    )),
    "col1,col2"
  )

  expect_error(parse_col_select(list(1:3)))
})

test_that("parse_col_select() removes duplicates and preserves order of first occurrence", {
  x <- c("A", "A", "B", "A", "C", "B")
  expect_identical(parse_col_select(x), "A,B,C")
})

test_that("errors on bad input", {
  expect_error(
    parse_col_select(1.0),
    "`col_select` must be a .+?character.+? vector, not a .+?numeric"
  )
  expect_error(
    parse_col_select(1L:3L),
    "`col_select` must be a .+?character.+? vector, not a .+?integer"
  )
  expect_error(
    parse_col_select(TRUE),
    "`col_select` must be a .+?character.+? vector, not a .+?logical"
  )
})

#' Get Open Data resources from a dataset
#'
#' @description Downloads multiple resources from a dataset on the NHS Open Data
#'  platform by dataset name, with optional row limits and context columns.
#'
#' @param dataset_name Name of the dataset as found on the
#' [NHS Open Data platform](https://www.opendata.nhs.scot) (character).
#' @param max_resources (optional) The maximum number of resources to return
#' (integer). If not set, all resources are returned.
#' @inheritParams get_resource
#'
#' @seealso [get_resource()] for downloading a single resource from a dataset.
#'
#' @return A [tibble][tibble::tibble-package] with the data.
#' @export
#'
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' \dontrun{
#' get_dataset("gp-practice-populations", max_resources = 2, rows = 10)
#' }
get_dataset <- function(
  dataset_name,
  max_resources = NULL,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = FALSE
) {
  # throw error if name type/format is invalid
  check_dataset_name(dataset_name)

  # define query and try API call
  query <- list(id = dataset_name)
  content <- try(
    phs_GET("package_show", query),
    silent = TRUE
  )

  # if content contains a 'Not Found Error'
  # throw error with suggested dataset name
  if (grepl("Not Found Error", content[1L], fixed = TRUE)) {
    suggest_dataset_name(dataset_name)
  }

  # define list of resource IDs to get
  all_ids <- purrr::map_chr(content$result$resources, ~ .x$id)

  n_res <- length(all_ids)
  res_index <- 1L:min(n_res, max_resources)

  selection_ids <- all_ids[res_index]

  # get all resources
  all_data <- purrr::map(
    selection_ids,
    get_resource,
    rows = rows,
    row_filters = row_filters,
    col_select = col_select
  )

  # resolve class issues
  types <- purrr::map(
    all_data,
    purrr::map_chr,
    class
  )

  # for each df, check if next df class matches
  inconsistencies <- vector(length = length(types) - 1L, mode = "list")
  for (i in seq_along(types)) {
    if (i == length(types)) break

    this_types <- types[[i]]
    next_types <- types[[i + 1L]]

    # find matching names
    matching_names <- suppressWarnings(
      names(this_types) == names(next_types)
    )

    # of matching name cols, find if types match too
    inconsistent_index <- this_types[matching_names] !=
      next_types[matching_names]
    inconsistencies[[i]] <- this_types[matching_names][inconsistent_index]
  }

  # define which columns to coerce and warn
  to_coerce <- unique(names(unlist(inconsistencies)))

  if (length(to_coerce) > 0L) {
    cli::cli_warn(c(
      "Due to conflicts between column types across resources,
      the following {cli::qty(to_coerce)} column{?s} ha{?s/ve} been coerced to type character:",
      "{.val {to_coerce}}"
    ))

    all_data <- purrr::map(
      all_data,
      dplyr::mutate,
      dplyr::across(
        dplyr::any_of(to_coerce),
        as.character
      )
    )
  }

  if (include_context) {
    # Add the 'resource context' as columns to the data
    all_data <- purrr::pmap(
      list(
        data = all_data,
        id = selection_ids,
        name = purrr::map_chr(content$result$resources[res_index], ~ .x$name),
        created_date = purrr::map_chr(
          content$result$resources[res_index],
          ~ .x$created
        ),
        modified_date = purrr::map_chr(
          content$result$resources[res_index],
          ~ .x$last_modified
        )
      ),
      add_context
    )
  }

  # Combine the list of resources into a single tibble
  combined <- purrr::list_rbind(all_data)

  return(combined)
}

test_that("Returned context is the same for resource and dataset", {
  skip_if_offline(host = "www.opendata.nhs.scot")

  dataset <- get_dataset(
    "general-practitioner-contact-details",
    rows = 10L,
    include_context = TRUE
  )

  res_id_1 <- "647b256e-4a03-4963-8402-bf559c9e2fff"
  resource_1 <- get_resource(
    res_id = res_id_1,
    rows = 10L,
    include_context = TRUE
  )

  res_id_2 <- "e37c14fe-51f7-4935-87d1-c79b30fe8824"
  resource_2 <- get_resource(
    res_id = res_id_2,
    rows = 10L,
    include_context = TRUE
  )

  expect_identical(
    dataset %>%
      dplyr::filter(ResID == res_id_1) %>%
      dplyr::select(!dplyr::where(~ anyNA(.x))),
    resource_1,
    # list_as_map = TRUE will sort variable names before comparing
    list_as_map = TRUE
  )
  expect_identical(
    dataset %>%
      dplyr::filter(ResID == res_id_2) %>%
      dplyr::select(!dplyr::where(~ anyNA(.x))),
    resource_2,
    list_as_map = TRUE
  )
})

test_that("add_context works with odd data (offline)", {
  df <- tibble::tibble(
    Surname = letters[1:5],
    Sex = rep("Female", 5L),
    Postcode = rep("EH1 1AA", 5L),
    HB = rep("S08000030", 5L),
    HSCP = rep("S37000026", 5L)
  )

  res_id <- "647b256e-4a03-4963-8402-bf559c9e2fff"
  data <- add_context(
    data = df,
    id = res_id,
    name = "GP Details April 2024",
    created_date = "2025-01-01T00:00:00",
    modified_date = NULL
  )

  expect_s3_class(data, "tbl_df")
  expect_named(data)
  expect_contains(
    names(data),
    c(
      "ResID",
      "ResName",
      "ResCreatedDate",
      "ResModifiedDate",
      "Surname",
      "Sex",
      "Postcode",
      "HB",
      "HSCP"
    )
  )
  expect_true(all(is.na(data$ResModifiedDate)))
})

test_that("add_context() coerces modified_date >= created_date (offline)", {
  # 'modified_date' is slightly earlier than 'created_date' — function should correct it
  df <- tibble::tibble(A = 1L:3L, B = letters[1:3])

  created <- "2025-01-01T00:00:00"
  modified <- "2024-12-31T23:59:59" # earlier than created

  out <- add_context(
    data = df,
    id = "abc-123",
    name = "Example",
    created_date = created,
    modified_date = modified
  )

  expect_s3_class(out, "tbl_df")
  expect_true(all(out$ResModifiedDate >= out$ResCreatedDate))
  expect_identical(nrow(out), nrow(df))
})

test_that("add_context() prepends context columns in fixed order (offline)", {
  df <- tibble::tibble(x = 1L:2L, y = 3L:4L)

  out <- add_context(
    data = df,
    id = "abc-123",
    name = "Example",
    created_date = "2025-01-01T00:00:00",
    modified_date = "2025-01-01T00:00:01"
  )

  expect_identical(
    names(out)[1:4],
    c("ResID", "ResName", "ResCreatedDate", "ResModifiedDate")
  )
  expect_identical(names(out)[-(1:4)], names(df))
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("get_dataset returns data in the expected format", {
  n_resources <- 2L
  n_rows <- 2L
  data <- get_dataset(
    dataset_name = "gp-practice-populations",
    max_resources = n_resources,
    rows = n_rows
  )

  expect_s3_class(data, "tbl_df")
  expect_identical(nrow(data), n_resources * n_rows)
  expect_named(data)
  expect_gte(ncol(data), 4L)
})

test_that("get_dataset works properly with filters", {
  n_resources <- 3L
  n_rows <- 10L
  columns <- c("Date", "PracticeCode", "HSCP", "AllAges")

  data <- get_dataset(
    "gp-practice-populations",
    max_resources = n_resources,
    rows = n_rows,
    row_filters = list(HSCP = "S37000026"),
    col_select = columns
  )

  expect_s3_class(data, "tbl_df")
  expect_identical(nrow(data), n_resources * n_rows)
  expect_named(data, columns)
  expect_true(all(data[["HSCP"]] == "S37000026"))
})

test_that("get_dataset errors properly", {
  expect_error(
    get_dataset("Mal-formed-name"),
    regexp = "The dataset name supplied `Mal-formed-name` is invalid"
  )
  expect_error(
    get_dataset("dataset-name-with-no-close-match"),
    regexp = "Can't find the dataset name `dataset-name-with-no-close-match`"
  )
  expect_error(
    get_dataset("gp-practice-population"),
    regexp = "Did you mean .+?gp-practice-populations.+?\\?"
  )
})

test_that("get_dataset filters error properly", {
  expect_error(
    get_dataset("gp-practice-populations", col_select = "Non-existent column"),
    regexp = "API error"
  )
})

test_that("get_dataset works with multiple filters", {
  n_resources <- 3L
  columns <- c("Date", "PracticeCode", "HSCP", "AllAges")

  data <- get_dataset(
    "gp-practice-populations",
    max_resources = n_resources,
    row_filters = list(PracticeCode = c("10002", "10017")),
    col_select = columns
  )

  expect_s3_class(data, "tbl_df")
  expect_gte(nrow(data), n_resources * 4L)
  expect_named(data, columns)
  expect_in(data[["PracticeCode"]], c("10002", "10017"))
})

test_that("Warns when having to coerce types", {
  expect_warning(
    coerced_data <- get_dataset(
      dataset_name = "nhsscotland-payments-to-general-practice",
      rows = 1L,
      col_select = "PracticeListSize"
    ),
    "Due to conflicts between column types across resources"
  )

  expect_s3_class(coerced_data, "tbl_df")
  expect_named(coerced_data, "PracticeListSize")
  expect_type(coerced_data[["PracticeListSize"]], "character")
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns expected context with the data", {
  n_rows <- 2L
  data_col_names <- names(
    get_dataset(
      dataset_name = "gp-practice-populations",
      max_resources = 1L,
      rows = 1L
    )
  )
  data <- get_dataset(
    dataset_name = "gp-practice-populations",
    max_resources = 1L,
    rows = n_rows,
    include_context = TRUE
  )

  expect_s3_class(data, "tbl_df")
  expect_type(data$ResID, "character")
  expect_type(data$ResName, "character")
  expect_s3_class(data$ResCreatedDate, "POSIXct")
  expect_s3_class(data$ResModifiedDate, "POSIXct")

  expect_identical(nrow(data), n_rows)
  expect_named(
    data,
    c("ResID", "ResName", "ResCreatedDate", "ResModifiedDate", data_col_names)
  )
})

test_that("returns expected context and length of data", {
  n_resources <- 2L
  n_rows <- 2L

  data <- get_dataset(
    dataset_name = "gp-practice-populations",
    max_resources = n_resources,
    rows = n_rows,
    include_context = TRUE
  )

  expect_s3_class(data, "tbl_df")
  expect_type(data$ResID, "character")
  expect_type(data$ResName, "character")
  expect_s3_class(data$ResCreatedDate, "POSIXct")
  expect_s3_class(data$ResModifiedDate, "POSIXct")

  expect_identical(nrow(data), n_resources * n_rows)
  expect_length(unique(data[["ResID"]]), n_resources)
})

#' get_latest_resource_id
#'
#' to be confident that the resource returned is the one intended
#' two conditions have to be met. It has to appear at the top of
#' of the resource list as shown on the open data platform.
#' The order they are returned via the api is the same
#' as they appear on the open data platform. It also
#' has to have the most recent date created
#'
#' There are only some datasets that this functionality
#' is relevant to, these are listed within applicable
#' datasets and are the datasets that keep historic
#' resources instead of over writing them.
#'
#' @inheritParams get_dataset
#'
#' @return a string with the resource id
#' @keywords internal
#' @noRd
get_latest_resource_id <- function(dataset_name, call = rlang::caller_env()) {
  # send the api request
  query <- list(id = dataset_name)
  content <- phs_GET("package_show", query)

  # add the id, created date and last_modified to a data.frame
  id <- vector("character")
  created_date <- vector("character")
  modified_date <- vector("character")

  for (res in content$result$resources) {
    id <- append(id, res$id)
    created_date <- append(created_date, res$created)
    modified_date <- append(modified_date, res$last_modified)
  }
  all_id_data <- tibble::tibble(
    id = id,
    created_date = strptime(created_date, format = "%FT%X", tz = "UTC"),
    modified_date = strptime(modified_date, format = "%FT%X", tz = "UTC")
  ) %>%
    dplyr::mutate(most_recent_date_created = max(created_date))

  # get the first row of the resources, this will be the same that appears
  # on the top on the open data platform
  all_id_data_first_row <- dplyr::slice_head(all_id_data, n = 1L)

  # If the resource at the top as appearing on the open data platform also has the most
  # recent date created, return it. Otherwise, error
  if (
    all_id_data_first_row$created_date ==
      all_id_data_first_row$most_recent_date_created
  ) {
    return(all_id_data_first_row$id)
  }
  cli::cli_abort("The most recent id could not be identified", call = call)
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns data for a dataset that is listed", {
  expect_no_error(data <- get_latest_resource("gp-practice-populations"))
  expect_s3_class(data, "tbl_df")
  expect_match(data[["ResName"]], "^GP Practice Populations \\w+ [0-9]{4}$")
  expect_named(data)
  expect_contains(
    names(data),
    c(
      "ResID",
      "ResName",
      "ResCreatedDate",
      "ResModifiedDate",
      "Date",
      "HB",
      "HSCP",
      "Sex"
    )
  )
})

test_that("returns error for a dataset that is not listed", {
  expect_error(
    get_latest_resource("hospital-codes"),
    "not within the applicable datasets"
  )
})

test_that("errors for ambiguous resource", {
  expect_error(
    get_latest_resource_id("weekly-covid-19-statistical-data-in-scotland"),
    "The most recent id could not be identified"
  )
})

#' Throw an error with suggested dataset name, if possible.
#'
#' @inheritParams get_dataset
#' @keywords internal
#' @noRd
suggest_dataset_name <- function(dataset_name, call = rlang::caller_env()) {
  content <- phs_GET("package_list", "")

  dataset_names <- unlist(content$result)

  # calculate string distances
  string_distances <- stringdist::stringdist(dataset_name, dataset_names)

  # if min distance is too big, abort
  if (min(string_distances) > 10.0) {
    cli::cli_abort(
      c(
        "Can't find the dataset name
      {.var {dataset_name}}, or a close match.",
        i = "Find a dataset's name in the URL
      of its page on {.url www.opendata.nhs.scot.}"
      ),
      call = call
    )
  }

  # find closet match
  closest_match <- dataset_names[which(
    string_distances == min(string_distances)
  )]

  # throw error with suggestion
  cli::cli_abort(
    c(
      "Can't find the dataset name {.var {dataset_name}}.",
      i = "Did you mean {?any of }{.val {closest_match}}?"
    ),
    call = call
  )
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("throws error and doesn't suggest for large distance", {
  expect_error(
    suggest_dataset_name(
      "135987645892erhusidhnjsfdhf92ry9823hr2iuh2eiyrhwfue"
    ),
    regexp = "Can't find the dataset name"
  )
})

test_that("throws error and does suggest for close matches", {
  expect_error(
    suggest_dataset_name(
      "rospital-codes"
    ),
    regexp = "Did you mean .+?\\?"
  )

  expect_error(
    suggest_dataset_name(
      "population"
    ),
    regexp = "Did you mean any of .+?\\?"
  )
})

#' Check if a resource ID is valid
#'
#' @description
#' Used to attempt to validate a res_id before submitting it to the API
#'
#' @inheritParams get_resource
#'
#' @return TRUE / FALSE indicating the validity of the res_id
#' @keywords internal
#' @noRd
check_res_id <- function(res_id, call = rlang::caller_env()) {
  # check res_id is single value
  if (length(res_id) > 1L) {
    cli::cli_abort(
      c(
        "Argument {.var res_id} must be of length 1.",
        i = "You supplied a res_id with a length of {length(res_id)}",
        x = "`get_resource` does not currently support
      requests for multiple resources simultaneously."
      ),
      call = call
    )
  }

  # check res_id is character
  if (!inherits(res_id, "character")) {
    cli::cli_abort(
      c(
        "Argument {.var res_id} must be of type character",
        i = "You supplied a {.var res_id} with type  {.cls {class(res_id)[1]}}"
      ),
      call = call
    )
  }

  # check regex pattern
  res_id_regex <-
    "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
  if (!grepl(res_id_regex, res_id)) {
    cli::cli_abort(
      c(
        "Argument {.var res_id} is in an invalid format.",
        i = "You can find a resource's ID in the URL of it's page on {.url www.opendata.nhs.scot}."
      ),
      call = call
    )
  }
}

test_that("errors for length > 1", {
  expect_error(
    check_res_id(letters),
    regexp = "You supplied a res_id with a length of 26"
  )
})

test_that("errors for incorrect type", {
  expect_error(
    check_res_id(20),
    regexp = "(must be of type character)"
  )
})

test_that("errors for invalid format", {
  expect_error(
    check_res_id("wrong format"),
    regexp = "is in an invalid format."
  )
})

test_that("returns nothing for correct format/length/type", {
  expect_null(
    check_res_id("a965ee86-0974-4c93-bbea-e839e27d7085")
  )
})

#' Get PHS Open Data using SQL
#'
#' @description
#' Downloads data from the NHS Open Data platform using a SQL query. Similar to
#' [get_resource()], but allows more flexible server-side querying. This
#' function has a lower maximum row number (32,000 vs 99,999) for returned
#' results.
#'
#' @param sql A single `SELECT` query (character).
#'
#' @details
#' Only 32,000 rows can be returned from a single SQL query.
#'
#' @section SQL syntax:
#' The resource ID must be double-quoted, e.g.
#' `SELECT * FROM "58527343-a930-4058-bf9e-3c6e5cb04010"`, as must column names,
#' e.g. `"Year"`. Strings require single quotes, e.g. `'value'`. This syntax is
#' needed because the CKAN DataStore uses
#' [PostgreSQL](https://www.postgresql.org/docs/9.2/index.html).
#'
#' @section R raw strings:
#' Enclosing the query in an R raw string avoids the need to escape embedded
#' quotes, e.g. `\"TotalCancelled\"`. Square brackets are the recommended
#' delimiter, i.e. `r"[...]"`, because `)"` within a query, e.g.
#' `SUM("TotalCancelled")`, would prematurely close the string. Another option
#' is `r"{...}"`, in the rare case that the query contains `]"`.
#'
#' @return
#' A [tibble][tibble::tibble-package] with the query results. Only 32,000 rows
#' can be returned from a single SQL query.
#'
#' @seealso
#' [get_resource()] for downloading a resource without using a SQL query.
#'
#' @export
#'
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' # Basic query
#' cancelled_ops <- get_resource_sql(r"[
#' SELECT
#'     "TotalCancelled",
#'     "TotalOperations",
#'     "Hospital",
#'     "Month"
#' FROM
#'     "bcc860a4-49f4-4232-a76b-f559cf6eb885"
#' WHERE
#'     "Hospital" = 'D102H'
#' ]")
#'
#' # Joining two resources
#' hb_pop <- get_resource_sql(r"[
#' SELECT
#'     pops."Year",
#'     pops."HB",
#'     lookup."HBName",
#'     pops."AllAges"
#' FROM
#'     "27a72cc8-d6d8-430c-8b4f-3109a9ceadb1" AS pops
#'     JOIN "652ff726-e676-4a20-abda-435b98dd7bdc" AS lookup ON pops."HB" = lookup."HB"
#' WHERE
#'     pops."Sex" = 'All'
#'     AND pops."Year" > 2006
#' ]")
get_resource_sql <- function(sql) {
  if (length(sql) != 1L) {
    cli::cli_abort(c(
      x = "SQL validation error.",
      i = "{.var sql} must be length 1 not {length(sql)}."
    ))
  }

  if (!inherits(sql, "character")) {
    cli::cli_abort(c(
      x = "SQL validation error.",
      i = "{.var sql} must be of class {.cls character} not {.cls {class(sql)}}."
    ))
  }

  # check query is a SELECT statement
  if (!grepl("^\\s*?SELECT", sql)) {
    cli::cli_abort(c(
      x = "SQL validation error.",
      i = "{.var sql} must start with {.val SELECT}"
    ))
  }

  # Add the SQL statement to the query
  query <- list(sql = sql)

  # attempt get request
  content <- phs_GET("datastore_search_sql", query)

  if (!is.null(content[["result"]][["records_truncated"]])) {
    cli::cli_warn(
      "The data was truncated because your query matched more than the
      maximum number of rows."
    )
  }

  # extract the records (rows) from content
  query_data <- purrr::map(
    content$result$records,
    ~ {
      # replace NULL with "" so tibble works
      is_null <- purrr::map_lgl(.x, is.null)
      .x[is_null] <- ""

      tibble::as_tibble(.x)
    }
  ) %>%
    purrr::list_rbind()

  # If the query returned no rows, exit now.
  if (nrow(query_data) == 0L) {
    return(query_data)
  }

  # get correct order of columns
  col_order <- purrr::map_chr(
    content$result$fields,
    ~ .x$id
  )
  col_order <- col_order[!col_order %in% c("_id", "_full_text")]

  # select and reorder columns to reflect
  cleaner <- dplyr::select(query_data, dplyr::all_of(col_order))

  # warn if limit may have been surpassed
  if (nrow(cleaner) == 32000L) {
    cli::cli_warn(c(
      "Row number limit",
      i = "SQL queries are limitted to returning 32,000 results.
      This may have affected the results of your query."
    ))
  }

  return(cleaner)
}

test_that("throws errors on invalid sql argument", {
  # wrong class
  expect_error(
    get_resource_sql(9000),
    regexp = "must be of class"
  )

  # wrong length
  expect_error(
    get_resource_sql(letters),
    regexp = "must be length 1 not 26\\."
  )

  # wrong start
  expect_error(
    get_resource_sql("this is wrong"),
    regexp = "`sql` must start with"
  )
})

test_that("gets expected data for a simple SQL query", {
  skip_if_offline(host = "www.opendata.nhs.scot")
  data <- get_resource_sql(
    sql = "
     SELECT
       \"TotalCancelled\",\"TotalOperations\",\"Hospital\",\"Month\"
     FROM
       \"bcc860a4-49f4-4232-a76b-f559cf6eb885\"
     WHERE
       \"Hospital\" = 'D102H'
  "
  )

  expect_s3_class(data, "tbl")
  expect_identical(unique(data$Hospital), "D102H")
  expect_named(
    data,
    c("TotalCancelled", "TotalOperations", "Hospital", "Month")
  )
})

test_that("gets expected data for a joined SQL query", {
  skip_if_offline(host = "www.opendata.nhs.scot")
  data <- get_resource_sql(
    sql = paste(
      "SELECT pops.\"Year\", pops.\"HB\", lookup.\"HBName\", pops.\"AllAges\"",
      "FROM \"27a72cc8-d6d8-430c-8b4f-3109a9ceadb1\" AS pops",
      "JOIN \"652ff726-e676-4a20-abda-435b98dd7bdc\" AS lookup",
      "ON pops.\"HB\" = lookup.\"HB\"",
      "WHERE pops.\"Sex\" = 'All' AND pops.\"Year\" > 2006"
    )
  )

  expect_s3_class(data, "tbl")
  expect_gt(min(as.integer(data$Year)), 2006L)
  expect_named(data, c("Year", "HB", "HBName", "AllAges"))
})

test_that("SQL errors", {
  skip_if_offline(host = "www.opendata.nhs.scot")
  # non-existent column in real table
  expect_error(
    get_resource_sql(
      "SELECT \"Hospital\",\"donut\"
      from \"bcc860a4-49f4-4232-a76b-f559cf6eb885\""
    ),
    regexp = "column \"donut\" does not exist"
  )

  # non-existent table
  expect_error(
    get_resource_sql(
      "SELECT * from \"donut\""
    ),
    regexp = "resource/table \"donut\" does not exist"
  )

  # syntax errors
  expect_error(
    get_resource_sql(
      "SELECT * from 'donut'"
    ),
    regexp = "syntax error at or near \"'donut'\""
  )
})

#' Get Open Data resource
#'
#' @description Downloads a single resource from the NHS Open Data platform by resource ID, with optional filtering and column selection.
#'
#' @param res_id The resource ID as found on \href{https://www.opendata.nhs.scot/}{NHS Open Data platform} (character).
#' @param rows (optional) Maximum number of rows to return (integer).
#' @param row_filters (optional) A named list or vector specifying values of columns/fields to keep (e.g., list(Date = 20220216, Sex = "Female")).
#' @param col_select (optional) A character vector containing the names of desired columns/fields (e.g., c("Date", "Sex")).
#' @param include_context (optional) If `TRUE`, additional information about the resource will be added as columns to the data, including the resource ID, the resource name, the creation date, and the last modified/updated date.
#'
#' @seealso [get_dataset()] for downloading all resources from a given dataset.
#'
#' @return A [tibble][tibble::tibble-package] with the data.
#' @export
#'
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' res_id <- "ca3f8e44-9a84-43d6-819c-a880b23bd278"
#'
#' data <- get_resource(res_id)
#'
#' filters <- list("HB" = "S08000030", "Month" = "202109")
#' wanted_cols <- c("HB", "Month", "TotalPatientsSeen")
#'
#' filtered_data <- get_resource(
#'   res_id = res_id,
#'   row_filters = filters,
#'   col_select = wanted_cols
#' )
get_resource <- function(
  res_id,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = FALSE
) {
  # check res_id
  check_res_id(res_id)

  parsed_col_select <- parse_col_select(col_select)
  parsed_row_filters <- parse_row_filters(row_filters)

  if (
    is.logical(parsed_row_filters) &&
      !parsed_row_filters &&
      !is.null(row_filters)
  ) {
    col_select_sql <- dplyr::if_else(
      is.null(col_select),
      "*",
      paste0("\"", paste(col_select, collapse = "\",\""), "\"")
    )

    row_filters_sql <- paste(
      purrr::imap_chr(
        row_filters,
        function(value, col) {
          paste0("\"", col, "\"=\'", value, "\'", collapse = " OR ")
        }
      ),
      collapse = ") AND ("
    )

    sql <- sprintf(
      "SELECT %s FROM \"%s\" WHERE (%s) %s",
      col_select_sql,
      res_id,
      row_filters_sql,
      dplyr::if_else(is.null(rows), "", paste("LIMIT", rows))
    )

    return(get_resource_sql(sql))
  }

  # define query
  query <- list(
    id = res_id,
    limit = rows,
    q = parsed_row_filters,
    fields = parsed_col_select
  )

  # if dump should be used, use it
  if (use_dump_check(query, rows)) {
    data <- dump_download(res_id)
  } else {
    # if there is no row limit set
    # set limit to CKAN max
    if (is.null(query$limit)) query$limit <- 99999L

    # remove null values from query
    query <- purrr::compact(query)

    # fetch the data
    res_content <- phs_GET("datastore_search", query)

    # if the total number of rows is greater than the
    # number of rows fetched
    # AND the user was not aware of this limit (`rows` defaulted to NULL)
    # warn the user about this limit.
    total_rows <- res_content$result$total
    if (is.null(rows) && query$limit < total_rows) {
      cli::cli_warn(c(
        "Returning the first {query$limit}
      results (rows) of your query.
      {total_rows} rows match your query in total.",
        i = "To get ALL matching rows you will need to download
      the whole resource and apply filters/selections locally."
      ))
    }

    # if more rows were requested than received
    # let the user know
    if (!is.null(rows) && query$limit > total_rows) {
      cli::cli_warn(
        "You set {.var rows} to {.val {query$limit}} but only {.val {total_rows}} rows matched your query."
      )
    }

    # extract data from response content
    data <- purrr::map(
      res_content$result$records,
      ~.x
    ) %>%
      dplyr::bind_rows() %>%
      dplyr::select(
        -dplyr::starts_with("rank "),
        -dplyr::matches("_id")
      )
  }

  if (include_context) {
    # Get resource context if required
    context_content <- phs_GET(
      action = "resource_show",
      query = paste0("id=", res_id)
    )

    res_id <- context_content$result$id
    res_name <- context_content$result$name
    res_created_date <- context_content$result$created
    res_modified_date <- context_content$result$last_modified

    data <- add_context(
      data = data,
      id = res_id,
      name = res_name,
      created_date = res_created_date,
      modified_date = res_modified_date
    )
  }

  return(data)
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns expected context with the data", {
  gp_list_apr_2021 <- "a794d603-95ab-4309-8c92-b48970478c14"
  data_col_names <- names(get_resource(res_id = gp_list_apr_2021, rows = 1L))

  # without query
  data <- get_resource(
    res_id = gp_list_apr_2021,
    rows = 10L,
    include_context = TRUE
  )

  expect_s3_class(data, "tbl_df")
  expect_type(data$ResID, "character")
  expect_type(data$ResName, "character")
  expect_s3_class(data$ResCreatedDate, "POSIXct")
  expect_s3_class(data$ResModifiedDate, "POSIXct")

  expect_length(data, 19L)
  expect_identical(nrow(data), 10L)
  expect_named(
    data,
    c(
      "ResID",
      "ResName",
      "ResCreatedDate",
      "ResModifiedDate",
      data_col_names
    )
  )

  # with query
  data_q <- get_resource(
    gp_list_apr_2021,
    row_filters = list(PracticeCode = 10002),
    col_select = c("PracticeCode", "AddressLine1"),
    include_context = TRUE
  )

  expect_named(
    data_q,
    c(
      "ResID",
      "ResName",
      "ResCreatedDate",
      "ResModifiedDate",
      "PracticeCode",
      "AddressLine1"
    )
  )
  expect_identical(data_q[["PracticeCode"]], 10002L)
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns full data if only res_id is input", {
  gp_list_apr_2021 <- "a794d603-95ab-4309-8c92-b48970478c14"

  data <- get_resource(res_id = gp_list_apr_2021)

  expect_identical(nrow(data), 926L)
})

test_that("returns full data if rows is set to over 99999", {
  gp_list_apr_2021 <- "a794d603-95ab-4309-8c92-b48970478c14"

  expect_warning(
    data <- get_resource(
      res_id = gp_list_apr_2021,
      rows = 9999999,
      row_filters = list(GPPracticeName = "The Blue Practice")
    ),
    regexp = "Can't request over 99,999 rows"
  )

  expect_identical(nrow(data), 926L)
})

test_that("first 99999 rows returned if query matches > 99999 rows", {
  prescriptions_apr_2021 <- "51b7ad3f-6d52-4165-94f4-92e322656c85"

  expect_warning(
    df <- get_resource(prescriptions_apr_2021, col_select = "HBT"),
    regexp = "(Returning the first 99999 results)"
  )

  expect_identical(nrow(df), 99999L)
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns data in the expected format", {
  gp_list_apr_2021 <- "a794d603-95ab-4309-8c92-b48970478c14"

  # without query
  data <- get_resource(res_id = gp_list_apr_2021, rows = 1L)

  expect_s3_class(data, "tbl_df")
  expect_gte(ncol(data), 8L)
  expect_identical(nrow(data), 1L)
  expect_named(data)

  # with query
  data_q <- get_resource(
    gp_list_apr_2021,
    row_filters = list(PracticeCode = 10002),
    col_select = c("PracticeCode", "AddressLine1")
  )

  expect_named(data_q, c("PracticeCode", "AddressLine1"))
  expect_identical(data_q[["PracticeCode"]], 10002L)

  expect_no_warning(get_resource("3e86b6fb-2062-4f05-8f4d-0bb001155d64"))
})

test_that("returns data with row specifications", {
  gp_list_apr_2021 <- "a794d603-95ab-4309-8c92-b48970478c14"

  expect_identical(
    nrow(get_resource(res_id = gp_list_apr_2021, rows = 926L)),
    926L
  )

  expect_identical(
    nrow(get_resource(res_id = gp_list_apr_2021, rows = 999L)),
    926L
  ) %>%
    expect_warning()
})

test_that("returns data for multiple filters", {
  data_row_limit <- get_resource(
    res_id = "e4985a62-9d59-4e71-8800-3f7ca29ffe0c",
    col_select = c("GPPractice", "DMDCode"),
    row_filters = list(GPPractice = c("80005", "80202")),
    rows = 100L
  )

  expect_s3_class(data_row_limit, "tbl_df")
  expect_identical(nrow(data_row_limit), 100L)
  expect_named(data_row_limit, c("GPPractice", "DMDCode"))

  data_full <- get_resource(
    res_id = "e4985a62-9d59-4e71-8800-3f7ca29ffe0c",
    col_select = c("GPPractice", "DMDCode", "PrescribedType"),
    row_filters = list(
      GPPractice = c("80005", "80202"),
      PrescribedType = "AMP"
    )
  )

  expect_s3_class(data_full, "tbl_df")
  expect_identical(nrow(data_full), 1108L)
  expect_named(data_full, c("GPPractice", "DMDCode", "PrescribedType"))
  expect_length(unique(data_full$GPPractice), 2L)
  expect_length(unique(data_full$PrescribedType), 1L)
})

test_that("returns data for multiple filters in mixed format", {
  delays <- get_resource(
    res_id = "fd354e4b-6211-48ba-8e4f-8356a5ed4215",
    col_select = c("MonthOfDelay", "ReasonForDelay", "NumberOfDelayedBedDays"),
    row_filters = list(HBT = "S08000031", MonthOfDelay = 201607:201707)
  )

  expect_s3_class(delays, "tbl_df")
  expect_gte(nrow(delays), 13L)
  expect_named(
    delays,
    c("MonthOfDelay", "ReasonForDelay", "NumberOfDelayedBedDays")
  )
  expect_length(unique(delays$MonthOfDelay), 13L)
})

test_that("returns data for multiple filters for all columns", {
  prescriptions <- get_resource(
    res_id = "d1fbede3-98c4-436e-9e75-2ed807a36075",
    row_filters = list(
      HBT = "S08000015",
      DMDCode = c("940711000001101", "1004511000001101", "1014311000001109")
    )
  )

  expect_s3_class(prescriptions, "tbl_df")
  expect_identical(nrow(prescriptions), 114L)
  expect_named(
    prescriptions,
    c(
      "HBT",
      "GPPractice",
      "DMDCode",
      "BNFItemCode",
      "BNFItemDescription",
      "PrescribedType",
      "NumberOfPaidItems",
      "PaidQuantity",
      "GrossIngredientCost",
      "PaidDateMonth"
    )
  )
  expect_length(unique(prescriptions$GPPractice), 55L)
  expect_setequal(
    prescriptions$DMDCode,
    c("940711000001101", "1004511000001101", "1014311000001109")
  )
  expect_setequal(prescriptions$HBT, "S08000015")
})

test_that("errors on invalid filters", {
  # non-existent column in row_filters
  expect_error(
    delays <- get_resource(
      res_id = "fd354e4b-6211-48ba-8e4f-8356a5ed4215",
      col_select = c(
        "MonthOfDelay",
        "ReasonForDelay",
        "NumberOfDelayedBedDays"
      ),
      row_filters = c(HBT = "S08000031", Month = 201607)
    ),
    regexp = "row_filters: invalid value"
  )

  # non-existent column in col_select
  expect_error(
    delays <- get_resource(
      res_id = "fd354e4b-6211-48ba-8e4f-8356a5ed4215",
      col_select = c("Month", "ReasonForDelay", "NumberOfDelayedBedDays"),
      row_filters = c(HBT = "S08000031", MonthOfDelay = 201607)
    ),
    regexp = "col_select: invalid value"
  )
})

test_that("We can filter data with `Sex = 'All'`", {
  pops <- get_resource(
    res_id = "27a72cc8-d6d8-430c-8b4f-3109a9ceadb1",
    row_filters = list(Sex = "All"),
    col_select = c("Year", "HB", "AllAges", "Sex")
  )

  expect_s3_class(pops, "tbl_df")
  expect_identical(
    nrow(pops),
    length(unique(pops$Year)) * 15L
  )
  expect_named(
    pops,
    c(
      "Year",
      "HB",
      "AllAges",
      "Sex"
    )
  )
  expect_length(unique(pops$HB), 15L)
  expect_setequal(pops$Sex, "All")
})

#' Send a GET request to the PHS CKAN API
#'
#' @inheritParams request_url
#' @param verbose TRUE or FALSE. If TRUE, a success message will be printed to the console.
#' @return content of a httr::GET request
#' @keywords internal
#' @noRd
phs_GET <- function(
  action,
  query,
  verbose = FALSE,
  call = rlang::caller_env()
) {
  # define URL
  url <- request_url(action, query)

  # Attempt GET request, gently retrying up to 3 times
  response <- httr::RETRY(
    verb = "GET",
    url = url,
    terminate_on = 409L,
    user_agent = httr::user_agent(
      "phsopendata (https://github.com/Public-Health-Scotland/phsopendata)"
    )
  )

  # Check for a response from the server
  if (!inherits(response, "response")) {
    cli::cli_abort(
      c(
        "Can't connect to the CKAN server.",
        i = "Check your network or proxy settings."
      ),
      call = call
    )
  }

  # Extract the content from the HTTP response
  if (httr::http_type(response) %in% c("text/html", "application/json")) {
    content <- httr::content(response)
  } else if (httr::http_type(response) == "text/csv") {
    content <- readr::read_csv(
      file = I(httr::content(response, as = "text")),
      guess_max = Inf
    )
  } else {
    cli::cli_abort(
      "The response contained an unhandled content type: {httr::http_type(response)}"
    )
  }

  # detect/handle errors
  error_check(content, call = call)

  if (verbose) cat("GET request successful.\n")
  return(content)
}

test_that("returns httr::content", {
  skip_if_offline(host = "www.opendata.nhs.scot")

  content <- phs_GET("package_list", "")

  expect_true(content$success)

  expect_identical(
    content$help,
    "https://www.opendata.nhs.scot/api/3/action/help_show?name=package_list"
  )
})

test_that("error_check() works as expected", {
  skip_if_offline(host = "www.opendata.nhs.scot")

  # no error for valid endpoint
  expect_type(
    phs_GET("package_list", ""),
    "list"
  )

  # not found error
  expect_error(
    phs_GET("datastore_search", "id=doop"),
    regexp = 'Resource "doop" was not found.'
  )
})

test_that("request_url() works as expected (offline test)", {
  # invalid action argument
  expect_error(
    phs_GET("", ""),
    regexp = "API call failed"
  )
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns nothing if no error", {
  expect_null(
    error_check(phs_GET("package_list", ""))
  )
})

test_that("throws error if error in httr content", {
  expect_error(
    error_check(phs_GET("datastore_search", "id=doop")),
    regexp = 'Resource "doop" was not found.'
  )
})

#' Determines if the datastore dump should be used
#'
#' @param query a list of items to query
#' @param rows Integer(1) or NULL. Number of rows requested.
#'
#' @return TRUE to use the datastore dump; FALSE to use the standard endpoint.#' @keywords internal
#' @noRd
use_dump_check <- function(query, rows) {
  null_query <- purrr::every(list(query$q, query$filter, query$fields), is.null)
  null_rows <- is.null(rows)

  if (null_query && null_rows) {
    # No filters specified - Use dump
    return(TRUE)
  } else if (null_rows) {
    # Filters specified, but no row limit - Don't use dump
    return(FALSE)
  } else if (rows > 99999L) {
    if (null_query) {
      # No filters specified but rows over 99999 requested. - Use dump
      cli::cli_warn(c(
        "Getting all rows of resource.",
        i = "All rows will be returned if you
      request over 99,999 rows of data.",
        i = "You set {.var rows} to
      {format(rows, big.mark = ',', scientific = FALSE)}"
      ))
      return(TRUE)
    } else {
      # Filters specified and rows over 99999 requested. - Use dump
      cli::cli_warn(c(
        "Invalid combination of {.var rows}, {.var row_filters}
      and/or {.var col_select}.",
        x = "Can't request over 99,999 rows of a resource
      AND query its rows/columns.",
        i = "ALL rows and columns of the resource will be downloaded."
      ))
      return(TRUE)
    }
  }

  # Normal query with filters and/or small num of rows - Don't use dump
  return(FALSE)
}

test_that("returns true as expected", {
  # all are null
  expect_true(
    use_dump_check(list(), NULL)
  )

  # rows > 99999 and all query is NULL
  expect_true(
    use_dump_check(list(), 100000L)
  ) %>%
    expect_warning()

  # query entries are not all NULL and rows > 99999
  expect_true(
    use_dump_check(list(q = 4L), 100000L)
  ) %>%
    expect_warning()

  expect_true(
    use_dump_check(list(q = 4L), 100000L)
  ) %>%
    expect_warning()
})

test_that("returns false as expected", {
  # rows is NULL and query list is not all NULL
  expect_false(
    use_dump_check(list(fields = "Age"), NULL)
  )

  # rows is under 99999 and query list is not all NULL
  expect_false(
    use_dump_check(list(fields = "Age"), 100L)
  )

  # rows is under 99999 and query list is all NULL
  expect_false(
    use_dump_check(list(), 100L)
  )
})

#' get a datasets additional info
#'
#' `get_dataset_additional_info()` returns a tibble of dataset names along with
#' the amount of resources it has and the date it was last updated.Last updated
#' is taken to mean the most recent date a resource within the dataset was
#' created or modified.
#'
#' @inheritParams get_dataset
#'
#' @return a [tibble][tibble::tibble-package] with the data
#' @export
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' get_dataset_additional_info("gp-practice-populations")
get_dataset_additional_info <- function(dataset_name) {
  # define query

  query <- list(id = dataset_name)
  # fetch the data
  content <- phs_GET("package_show", query)

  # get the amount of resources
  amount_of_resources <- length(content$result$resources)

  # get the last recourse created and modified dates
  last_resource_created_date <- purrr::map_chr(
    content$result$resources,
    ~ .$created
  )
  last_resource_modified_date <- purrr::map_chr(
    content$result$resources,
    ~ .$last_modified
  )

  # get the latest between the created and modified dates and
  # change to datetime format
  most_recent_resource_date <- max(
    last_resource_modified_date,
    last_resource_created_date
  ) %>%
    as.POSIXct(format = "%FT%X", tz = "UTC")

  # create tibble to return
  return_value <- tibble::tibble(
    name = dataset_name,
    n_resources = amount_of_resources,
    last_updated = most_recent_resource_date
  )

  return(return_value)
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns data in the expected format", {
  dataset <- get_dataset_additional_info(
    "weekly-accident-and-emergency-activity-and-waiting-times"
  )

  expect_s3_class(dataset, "tbl_df")
  expect_identical(nrow(dataset), 1L)
  expect_named(dataset, c("name", "n_resources", "last_updated"))
})

#' Provides an overview of all resources available on the Open Data platform.
#'
#' @description Provides an overview of all resources available from
#'  [opendata.nhs.scot](https://www.opendata.nhs.scot/), with the option to limit results based on both dataset
#'  and resource names. The returned tibble can be used to look-up dataset
#'  and resource ids, and is useful for exploring the available data sets.
#'
#' @param dataset_contains A character string containing an expression to be
#' @param dataset_contains A character string containing an expression to be
#'  used as search criteria against the dataset name.
#' @param resource_contains A character string containing a [regular expression](https://www.geeksforgeeks.org/dsa/write-regular-expressions/)
#'  to be matched against available resource names.
#' @param dataset_name Deprecated. Use dataset_contains instead.
#'
#' @return A [tibble][tibble::tibble-package] containing details of all available datasets and
#'  resources, or those containing the string specified in the
#'  `dataset_contains` and `resource_contains` arguments.
#'
#' @examples
#'
#' list_resources()
#' list_resources(dataset_contains = "standard-populations")
#' list_resources(
#'   dataset_contains = "standard-populations", resource_contains = "European"
#' )
#'
#' @export
list_resources <- function(
  dataset_contains = NULL,
  resource_contains = NULL,
  dataset_name = lifecycle::deprecated()
) {
  # Handling any use of "list_resources(dataset_name)"
  if (lifecycle::is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "1.0.3",
      what = "list_resources(dataset_name)",
      with = "list_resources(dataset_contains)"
    )
    dataset_contains <- dataset_name
  }

  # Normalise empty/whitespace strings to NULL
  normalise <- function(string) {
    if (is.null(string)) {
      return(NULL)
    }
    if (!is.character(string)) {
      return(string)
    }
    if (length(string) != 1L) {
      return(string)
    }
    if (is.na(string)) {
      return(NULL)
    }
    string_trimmed <- trimws(string)
    if (identical(string_trimmed, "")) {
      return(NULL)
    }
    return(string_trimmed)
  }
  dataset_contains <- normalise(dataset_contains)
  resource_contains <- normalise(resource_contains)

  # Validate that `dataset_contains` is NULL or a length-1 value
  if (!is.null(dataset_contains)) {
    if (!inherits(dataset_contains, "character")) {
      cli::cli_abort(c(
        "i" = "{.arg dataset_contains} must be {.class character} not a {.class {class(dataset_contains)}}."
      ))
    } else if (length(dataset_contains) != 1) {
      cli::cli_abort(c(
        "x" = "{.arg dataset_contains} must be {.val NULL} or length 1 not length {length(dataset_contains)}.",
        "i" = "Provide a single string (or leave it NULL) for this filter."
      ))
    }
  }

  # Validate that `resource_contains` is NULL or a length-1 value
  if (!is.null(resource_contains)) {
    if (!inherits(resource_contains, "character")) {
      cli::cli_abort(c(
        "i" = "{.arg resource_contains} must be {.class character} not a {.class {class(resource_contains)}}."
      ))
    } else if (length(resource_contains) != 1) {
      cli::cli_abort(c(
        "!" = "{.arg resource_contains} must be {.val NULL} or length 1 not length {length(resource_contains)}.",
        "i" = "Provide a single string (or leave it NULL) for this filter."
      ))
    }
  }

  validate_regex <- function(pattern, arg_name, call = rlang::caller_env()) {
    if (is.null(pattern)) {
      return(invisible(TRUE))
    }
    tryCatch(
      {
        suppressWarnings(grepl(pattern, "", perl = TRUE))
        TRUE
      },
      error = function(e) {
        cli::cli_abort(
          c(
            "i" = "{.arg {arg_name}} must be a valid regular expression.",
            "x" = e$message
          ),
          call = call
        )
      }
    )
  }

  validate_regex(resource_contains, "resource_contains")
  validate_regex(dataset_contains, "dataset_contains")

  data_tibble <- list_resources_query()

  build_and_anywhere_regex <- function(query) {
    # Split on one-or-more whitespace
    tokens <- strsplit(query, "\\s+")[[1]]
    tokens <- tokens[nzchar(tokens)]

    # Ensures all elements in 'tokens' appear somewhere in the string
    paste0("(?=.*", tokens, ")", collapse = "")
  }

  if (!is.null(resource_contains)) {
    resource_contains_regex <- build_and_anywhere_regex(resource_contains)

    data_tibble <- data_tibble[
      grepl(
        resource_contains_regex,
        data_tibble$resource_name,
        ignore.case = TRUE,
        perl = TRUE
      ),
    ]
    if (nrow(data_tibble) == 0) {
      cli::cli_warn(
        "No resources found for {.arg resource_contains} = {.val {resource_contains}}. Returning an empty tibble."
      )
    }
  }

  if (!is.null(dataset_contains)) {
    dataset_contains_regex <- build_and_anywhere_regex(dataset_contains)

    name_match <- grepl(
      dataset_contains_regex,
      data_tibble$dataset_name,
      ignore.case = TRUE,
      perl = TRUE
    )

    title_match <- grepl(
      dataset_contains_regex,
      data_tibble$dataset_title,
      ignore.case = TRUE,
      perl = TRUE
    )

    data_tibble <- data_tibble[name_match | title_match, ]

    if (nrow(data_tibble) == 0) {
      cli::cli_warn(
        "No resources found for the provided arguments. Returning an empty tibble."
      )
    }
  }

  return(data_tibble)
}


list_resources_query <- function() {
  # query for the API call
  query <- list(q = "*:*", rows = 3200)

  # API call
  content <- phs_GET(action = "package_search", query = query)

  # extract the data
  results <- purrr::chuck(content, "result", "results")

  datasets_list <- purrr::map(
    results,
    function(dataset) {
      resources <- purrr::chuck(dataset, "resources")

      tibble::tibble(
        resource_name = purrr::map_chr(
          resources,
          ~ purrr::pluck(.x, "name", .default = NA_character_)
        ),
        resource_id = purrr::map_chr(
          resources,
          ~ purrr::pluck(.x, "id", .default = NA_character_)
        ),
        dataset_name = purrr::chuck(dataset, "name"),
        dataset_title = purrr::chuck(dataset, "title"),
        dataset_id = purrr::chuck(dataset, "id"),
        url = purrr::map_chr(
          resources,
          ~ purrr::pluck(.x, "url", .default = NA_character_)
        ),
        last_modified = as.POSIXct(
          purrr::map_chr(
            resources,
            ~ purrr::pluck(.x, "last_modified", .default = NA_character_)
          ),
          format = "%FT%X",
          tz = "UTC"
        )
      )
    }
  )

  data_tibble <- purrr::list_rbind(datasets_list)

  return(data_tibble)
}

skip_if_offline(host = "www.opendata.nhs.scot")

all_data <- list_resources()

test_that("returns data in the expected format", {
  expect_s3_class(all_data, "tbl_df")
  expect_named(
    all_data,
    c(
      "resource_name",
      "resource_id",
      "dataset_name",
      "dataset_title",
      "dataset_id",
      "url",
      "last_modified"
    )
  )

  # Should have at least some rows and reasonable number of columns
  expect_gte(nrow(all_data), 1)
  expect_gte(ncol(all_data), 6)

  expect_type(all_data$resource_id, "character")
  expect_s3_class(all_data$last_modified, "POSIXct")
  expect_true(anyDuplicated(all_data$resource_id) == 0)
  expect_false(anyNA(all_data$resource_id))
  expect_false(anyNA(all_data$dataset_id))
})

test_that("filters by resource_contains (case-insensitive) and warns when empty", {
  # Pick a common term likely present across resources; adjust if needed
  data_eu <- list_resources(resource_contains = "european")
  expect_s3_class(data_eu, "tbl_df")
  expect_gte(nrow(data_eu), 0L)
  expect_match(
    object = data_eu$resource_name,
    regexp = "european",
    ignore.case = TRUE,
    all = TRUE
  )

  # Explicit empty case to assert warning + empty tibble
  expect_warning(
    data_none <- list_resources(
      resource_contains = "___no_such_resource___"
    ),
    regexp = "No resources found"
  )
  expect_equal(nrow(data_none), 0)
})

test_that("filters by dataset_contains (case-insensitive) and warns when empty", {
  data_pkg <- list_resources(dataset_contains = "hospital")
  expect_s3_class(data_pkg, "tbl_df")
  expect_gte(nrow(data_pkg), 0)

  name_ok <- grepl("hospital", data_pkg$dataset_name, ignore.case = TRUE)
  title_ok <- grepl("hospital", data_pkg$dataset_title, ignore.case = TRUE)

  expect_all_true(name_ok | title_ok)

  expect_warning(
    data_none <- list_resources(dataset_contains = "___no_such_package___"),
    regexp = "No resources found"
  )
  expect_identical(nrow(data_none), 0L)
})

test_that("combined filtering works", {
  # This test is intentionally permissive since the live catalogue changes.
  # Use broad terms that should commonly co-occur; tweak if it ever becomes too strict.
  data_combo <- list_resources(
    dataset_contains = "hospital",
    resource_contains = "admissions"
  )

  expect_s3_class(data_combo, "tbl_df")
  expect_named(
    data_combo,
    c(
      "resource_name",
      "resource_id",
      "dataset_name",
      "dataset_title",
      "dataset_id",
      "url",
      "last_modified"
    )
  )

  name_ok <- grepl("hospital", data_combo$dataset_name, ignore.case = TRUE)
  title_ok <- grepl("hospital", data_combo$dataset_title, ignore.case = TRUE)

  expect_all_true(name_ok | title_ok)

  expect_match(
    object = data_combo$resource_name,
    regexp = "admissions",
    ignore.case = TRUE,
    all = TRUE
  )
})

test_that("empty/whitespace/NA filters are treated as NULL", {
  expect_identical(
    object = list_resources(resource_contains = ""),
    expected = all_data
  )
  expect_identical(
    object = list_resources(resource_contains = "   "),
    expected = all_data
  )
  expect_identical(
    object = list_resources(dataset_contains = ""),
    expected = all_data
  )
  expect_identical(
    object = list_resources(dataset_contains = "   "),
    expected = all_data
  )
  expect_identical(
    object = list_resources(dataset_contains = NA_character_),
    expected = all_data
  )
  expect_identical(
    object = list_resources(resource_contains = NA_character_),
    expected = all_data
  )
})


test_that("invalid regex patterns error clearly", {
  expect_error(
    list_resources(resource_contains = "("),
    "must be a valid regular expression"
  )
  expect_error(
    list_resources(dataset_contains = "("),
    "must be a valid regular expression"
  )
})


test_that("input validation: dataset_contains/resource_contains must be NULL or length-1", {
  # Length > 1 should error
  expect_error(
    list_resources(dataset_contains = c("a", "b")),
    regexp = "length 1 not length 2"
  )

  # Non character should error
  expect_error(
    list_resources(dataset_contains = TRUE),
    "must be .*character"
  )
  expect_error(
    list_resources(resource_contains = 123),
    "must be .*character"
  )
})

test_that("dataset_name argument is deprecated", {
  expect_warning(
    list_resources(dataset_name = "hospital"),
    class = "lifecycle_warning_deprecated"
  )
})

test_that("dataset_contains is AND-anywhere (any order): 'Flu covid' == 'covid flu'", {
  res1 <- list_resources(dataset_contains = "Flu covid")
  res2 <- list_resources(dataset_contains = "covid flu")

  titles1 <- sort(unique(res1$dataset_title))
  titles2 <- sort(unique(res2$dataset_title))

  expect_gt(length(titles1), 0L)
  expect_identical(titles1, titles2)

  combined <- paste(res1$dataset_name, res1$dataset_title)
  expect_all_true(grepl("flu", combined, ignore.case = TRUE))
  expect_all_true(grepl("covid", combined, ignore.case = TRUE))
})

test_that("dataset_contains supports multi-term queries like 'GP list'", {
  res <- list_resources(dataset_contains = "GP list")
  titles <- unique(res$dataset_title)

  expect_gt(length(titles), 0L)
  expect_true(any(grepl(
    "GP Practice Contact Details and List Sizes",
    titles,
    fixed = TRUE
  )))

  combined <- paste(res$dataset_name, res$dataset_title)
  expect_all_true(grepl("gp", combined, ignore.case = TRUE))
  expect_all_true(grepl("list", combined, ignore.case = TRUE))
})

test_that("single-term dataset_contains behaves like a normal search", {
  res <- list_resources(dataset_contains = "GP")
  titles <- sort(unique(res$dataset_title))

  expect_true(any(grepl(
    "GP Practice Population Demographics",
    titles,
    fixed = TRUE
  )))
  expect_true(any(grepl(
    "GP Practice Contact Details and List Sizes",
    titles,
    fixed = TRUE
  )))

  combined <- paste(res$dataset_name, res$dataset_title)
  expect_all_true(grepl("gp", combined, ignore.case = TRUE))
})

test_that("dataset_contains allows regex (alternation) inside tokens", {
  # Should behave like 'Flu covid' because covid|covid is trivial,
  # but demonstrates regex parsing works.
  res <- list_resources(dataset_contains = "Flu (covid|covid)")

  combined <- paste(res$dataset_name, res$dataset_title)
  expect_all_true(grepl("flu", combined, ignore.case = TRUE))
  expect_all_true(grepl("covid", combined, ignore.case = TRUE))
})

test_that("resource_contains supports multi-term AND-anywhere semantics", {
  res <- list_resources(resource_contains = "GP sizes")
  skip_if(nrow(res) == 0)

  expect_all_true(grepl("gp", res$resource_name, ignore.case = TRUE))
  expect_all_true(grepl("sizes", res$resource_name, ignore.case = TRUE))
})

#' Get the latest resource from a data set
#'
#' Returns the latest resource available in a dataset.
#'
#' There are some datasets on the open data platform that
#' keep historic resources instead of updating existing ones.
#' For these it is useful to be able to retrieve the latest
#' resource. As of 1.8.2024 these data sets include:
#' * gp-practice-populations
#' * gp-practice-contact-details-and-list-sizes
#' * nhsscotland-payments-to-general-practice
#' * dental-practices-and-patient-registrations
#' * general-practitioner-contact-details
#' * prescribed-dispensed
#' * dispenser-location-contact-details
#' * community-pharmacy-contractor-activity
#'
#' @inheritParams get_dataset
#' @inheritParams get_resource
#'
#' @return a [tibble][tibble::tibble-package] with the data
#' @export
#'
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' \dontrun{
#' dataset_name <- "gp-practice-contact-details-and-list-sizes"
#'
#' data <- get_latest_resource(dataset_name)
#'
#' filters <- list("Postcode" = "DD11 1ES")
#' wanted_cols <- c("PracticeCode", "Postcode", "Dispensing")
#'
#' filtered_data <- get_latest_resource(
#'   dataset_name = dataset_name,
#'   row_filters = filters,
#'   col_select = wanted_cols
#' )
#' }
get_latest_resource <- function(
  dataset_name,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = TRUE
) {
  applicable_datasets <- c(
    "community-pharmacy-contractor-activity",
    "dental-practices-and-patient-registrations",
    "dispenser-location-contact-details",
    "general-practitioner-contact-details",
    "gp-practice-contact-details-and-list-sizes",
    "gp-practice-populations",
    "nhsscotland-payments-to-general-practice",
    "prescribed-dispensed"
  )

  # check if data set is within applicable datasets
  # throw error if not
  if (!dataset_name %in% applicable_datasets) {
    cli::cli_abort(
      c(
        "The dataset name supplied {.val {dataset_name}} is not within the applicable datasets.
      These are: {.val {applicable_datasets}}",
        x = "Please see {.fun get_latest_resource} documentation.",
        i = "You can find dataset names in the URL
      of a dataset's page on {.url www.opendata.nhs.scot}."
      ),
      call = rlang::caller_env()
    )
  }

  # get the latest resource id
  id <- get_latest_resource_id(dataset_name)

  resource_data <- get_resource(
    res_id = id,
    rows = rows,
    row_filters = row_filters,
    col_select = col_select,
    include_context = include_context
  )

  return(resource_data)
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns data for a dataset that is listed", {
  gp_latest <- get_latest_resource("gp-practice-populations")

  expect_s3_class(gp_latest, "tbl")
  expect_gt(nrow(gp_latest), 1L)
  expect_named(gp_latest)
  expect_contains(
    names(gp_latest),
    c("ResID", "ResName", "ResCreatedDate", "ResModifiedDate")
  )
  expect_s3_class(gp_latest$ResCreatedDate, "POSIXct")
  expect_s3_class(gp_latest$ResModifiedDate, "POSIXct")
})

test_that("returns error for a dataset that is not listed", {
  expect_error(
    get_latest_resource("hospital-codes"),
    "^The dataset name supplied.+?is not "
  )
})

#' Produces a URL for a GET request to opendata.nhs.scot
#'
#' @param action The API endpoint you want to use, e.g., "package_show" / "datastore_search".
#' @param query The query to pass to the endpoint defined by the action argument.
#' @return a URL as a character string
#' @keywords internal
#' @noRd
request_url <- function(action, query, call = rlang::caller_env()) {
  # check action is valid
  valid_actions <- c(
    "datastore_search",
    "datastore_search_sql",
    "dump",
    "package_show",
    "package_list",
    "resource_show",
    "package_search"
  )
  if (!(action %in% valid_actions)) {
    cli::cli_abort(
      c(
        "API call failed.",
        x = "{.val {action}} is an invalid {.arg action} argument."
      ),
      call = call
    )
  }

  base_url <- "https://www.opendata.nhs.scot"

  if (action == "dump") {
    # return dump URL
    url <- httr::modify_url(
      url = base_url,
      path = c("datastore/dump", query),
      query = list(bom = "true")
    )
  } else {
    url <- httr::modify_url(
      url = base_url,
      path = c("api/3/action", action),
      query = query
    )
  }

  # return standard API endpoint (i.e., not dump)
  return(url)
}

test_that("returns correct URL format", {
  expect_identical(
    request_url("datastore_search", "id=doop"),
    "https://www.opendata.nhs.scot/api/3/action/datastore_search?id=doop"
  )

  expect_identical(
    request_url("dump", "id=doop"),
    "https://www.opendata.nhs.scot/datastore/dump/id=doop?bom=true"
  )
})

test_that("request_url() builds URLs for remaining valid actions", {
  # package_show
  expect_identical(
    request_url("package_show", list(id = "gp-practice-populations")),
    "https://www.opendata.nhs.scot/api/3/action/package_show?id=gp-practice-populations"
  )
  # resource_show
  expect_identical(
    request_url(
      "resource_show",
      list(id = "a794d603-95ab-4309-8c92-b48970478c14")
    ),
    "https://www.opendata.nhs.scot/api/3/action/resource_show?id=a794d603-95ab-4309-8c92-b48970478c14"
  )
  # datastore_search_sql
  expect_identical(
    request_url("datastore_search_sql", list(sql = 'SELECT * FROM "xyz"')),
    "https://www.opendata.nhs.scot/api/3/action/datastore_search_sql?sql=SELECT%20%2A%20FROM%20%22xyz%22"
  )
})


test_that("rejects invalid actions", {
  expect_error(
    request_url("beep", ""),
    regexp = "API call failed."
  )
})

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("correctly extracts error from API response", {
  content <- httr::content(
    httr::GET(
      request_url("datastore_search", "id=doop")
    )
  )
  expect_identical(
    parse_error(content$error),
    "Not Found Error: Not found: Resource \"doop\" was not found."
  )

  content <- httr::content(
    httr::GET(
      request_url("datastore_search", "")
    )
  )
  expect_identical(
    parse_error(content$error),
    "resource_id: Missing value"
  )
})

#' Add resource context columns to a data frame
#'
#' @description Adds resource context columns (ID, name, created date, modified
#' date) to a data frame returned from a resource query.
#'
#' @param data A data frame or tibble containing resource data.
#' @param id The resource ID (character).
#' @param name The resource name (character).
#' @param created_date The resource creation date (character or POSIXct).
#' @param modified_date The resource last modified date (character or POSIXct).
#'
#' @return A data frame or tibble with context columns prepended.
#' @noRd
#' @keywords internal
add_context <- function(data, id, name, created_date, modified_date) {
  # Catch if the resource has never been modified
  if (is.null(modified_date)) {
    modified_date <- NA_character_
  }

  # Parse the date values
  created_date <- as.POSIXct(created_date, format = "%FT%X", tz = "UTC")
  modified_date <- as.POSIXct(modified_date, format = "%FT%X", tz = "UTC")

  # The platform can record the modified date as being before the created date
  # by a few microseconds, this will catch any rounding which ensure
  # created_date is always <= modified_date
  if (!is.na(modified_date) && modified_date < created_date) {
    modified_date <- created_date
  }

  data_with_context <- dplyr::mutate(
    data,
    ResID = id,
    ResName = name,
    ResCreatedDate = created_date,
    ResModifiedDate = modified_date,
    .before = dplyr::everything()
  )

  return(data_with_context)
}

#' Throws errors found in the content of an `httr::GET` request
#'
#' @param content object produced by `httr::content`
#' @keywords internal
#' @noRd
error_check <- function(content, call = rlang::caller_env()) {
  # if content is not a list,
  # stop for content (a string describing an error)
  if (!is.list(content)) {
    cli::cli_abort(
      c(
        "API error",
        x = content
      ),
      call = call
    )
  }

  # if there is no error status/message in the content,
  # break out of the function
  is_error <- suppressWarnings(
    !is.null(content$error)
  )
  if (!is_error) {
    return()
  }

  # generate error message and stop
  error_text <- parse_error(content$error)
  cli::cli_abort(
    c(
      "API error.",
      x = error_text
    ),
    call = call
  )
}

#' Create JSON 'dict' from named list or vector
#'
#' @description Formats a list or named vector into a valid query
#'
#' @param row_filters A named list or vector specifying values of columns/fields to keep (e.g., list(Date = 20220216, Sex = "Female")).
#'
#' @param row_filters list or named vectors matching fields to values
#' @return a json as a character string
#' @keywords internal
#' @noRd
parse_row_filters <- function(row_filters, call = rlang::caller_env()) {
  # exit function if no filters
  if (is.null(row_filters)) {
    return(NULL)
  }

  # Check if `row_filters` is a list or a character or numeric vector
  if (
    !rlang::is_bare_list(row_filters) &&
      !rlang::is_bare_character(row_filters) &&
      !rlang::is_bare_numeric(row_filters)
  ) {
    cli::cli_abort(
      "{.arg row_filters} must be a named {.cls list}, not a {.cls {class(row_filters)}}.",
      call = call
    )
  }

  # Ensure its elements are named
  if (is.null(names(row_filters)) || !all(nzchar(names(row_filters)))) {
    cli::cli_abort(
      "{.arg row_filters} should be a named {.cls list}.",
      call = call
    )
  }

  # check if any items in the list/vector are duplicates
  duplicates <- duplicated(names(row_filters))
  if (any(duplicates)) {
    cli::cli_abort(
      c(
        "Invalid input for {.arg row_filters}",
        x = "The {.val {names(row_filters)[which(duplicates)]}} filter{?s} {?is/are} duplicated.",
        i = "Only one filter per field is currently supported by {.fun get_resource}."
      ),
      call = call
    )
  }

  # Check if Sex = All was specified
  # There is a bug on CKAN which makes this not work, unless we use the SQL endpoint
  if ("Sex" %in% names(row_filters) && "All" %in% row_filters[["Sex"]]) {
    return(FALSE)
  }

  # check if any filters in list have length > 1
  multiple <- purrr::map_lgl(row_filters, ~ length(.x) > 1L)

  if (any(multiple)) {
    # Default to using SQL
    return(FALSE)
  }

  filter_body <- paste0(
    '"',
    names(row_filters),
    '":"',
    row_filters,
    '"',
    collapse = ","
  )

  return(paste0("{", filter_body, "}"))
}

test_that("returns NULL if `row_filters` = NULL", {
  expect_null(
    parse_row_filters(NULL)
  )
})

test_that("throws error for bad types", {
  expect_error(
    parse_row_filters(mtcars),
    regexp = " must be a named .+? not a "
  )
  expect_error(
    parse_row_filters(c(x = TRUE)),
    regexp = " must be a named .+? not a "
  )
  expect_error(
    parse_row_filters(c(x = NA)),
    regexp = " must be a named .+? not a "
  )
  expect_error(
    parse_row_filters(c(x = as.Date("2000-01-01"))),
    regexp = " must be a named .+? not a "
  )
})

test_that("returns FALSE for length > 1", {
  expect_false(
    parse_row_filters(list(x = letters))
  )
})

test_that("returns FALSE if `Sex = 'All'` is given", {
  expect_false(
    parse_row_filters(list(Sex = "All"))
  )

  expect_false(
    parse_row_filters(list(Sex = c("Male", "All")))
  )

  expect_false(
    parse_row_filters(list(
      Sex = "All",
      HBT = "S1234"
    ))
  )
})

test_that("throws error when un-named", {
  expect_error(
    parse_row_filters(list(1, 2)),
    regexp = " should be a named "
  )
  expect_error(
    parse_row_filters(list(a = 1, 2)),
    regexp = " should be a named "
  )
  expect_error(
    parse_row_filters(c(1, 2)),
    regexp = " should be a named "
  )
  expect_error(
    parse_row_filters(c(a = 1, 2)),
    regexp = " should be a named "
  )
})

test_that("throws error for non-unique names", {
  expect_error(
    parse_row_filters(list(x = 1, x = 2)),
    regexp = "Only one filter per field is currently supported"
  )
  expect_error(
    parse_row_filters(c(x = 1, x = 2)),
    regexp = "Only one filter per field is currently supported"
  )
})


test_that("returns JSON string from a named vector", {
  expect_true(
    jsonlite::validate(
      parse_row_filters(c(x = 5.0, y = 6.0))
    )
  )
  expect_true(
    jsonlite::validate(
      parse_row_filters(c(x = 5L, y = 6L))
    )
  )
  expect_true(
    jsonlite::validate(
      parse_row_filters(c(x = "a", y = "b"))
    )
  )
})

test_that("returns JSON string from list", {
  expect_true(
    jsonlite::validate(
      parse_row_filters(list(x = 5.0, y = 6.0))
    )
  )
  expect_true(
    jsonlite::validate(
      parse_row_filters(list(x = 5L, y = 6L))
    )
  )
  expect_true(
    jsonlite::validate(
      parse_row_filters(list(x = "a", y = "b"))
    )
  )
})

#' Throws an error if a dataset name is invalid
#'
#' @description
#' Used to attempt to validate a dataset name before submitting it to the API
#'
#' @param dataset_name a resource ID
#' @keywords internal
#' @noRd
check_dataset_name <- function(dataset_name, call = rlang::caller_env()) {
  # Starts and ends in a lowercase letter or number
  # Has only lowercase alphanum or hyphens inbetween
  dataset_name_regex <- "^[a-z0-9][a-z0-9\\-]+?[a-z0-9]$"

  if (!inherits(dataset_name, "character")) {
    cli::cli_abort(
      c(
        "The dataset name supplied {.var {dataset_name}} is invalid.",
        x = "dataset_name must be of type character.",
        i = "You supplied a {.cls {class(dataset_name)[0]}} value."
      ),
      call = call
    )
  }

  if (!grepl(dataset_name_regex, dataset_name)) {
    cli::cli_abort(
      c(
        "The dataset name supplied {.var {dataset_name}} is invalid",
        x = "dataset_name must be in dash-case
      (e.g., lowercase-words-separated-by-dashes).",
        i = "You can find dataset names in the URL
      of a dataset's page on {.url www.opendata.nhs.scot}."
      ),
      call = call
    )
  }
}

test_that("errors for incorrect case", {
  expect_error(
    check_dataset_name("hospital codes"),
    regexp = "dataset_name must be in dash-case"
  )
})

test_that("errors for incorrect type", {
  expect_error(
    check_dataset_name(20),
    regexp = "dataset_name must be of type character"
  )
})

test_that("returns nothing for valid type and format", {
  expect_null(
    check_dataset_name("hospital-codes")
  )
})

#' Lists all available datasets
#'
#' @description
#' `list_datasets()` shows all of the datasets hosted on the PHS Open Data
#' Platform.
#'
#' @section Lifecycle:
#'
#'
#' `r lifecycle::badge("superseded")`
#'
#' `list_datasets()` has been superseded by `list_resources()`.
#' While `list_datasets()` only returns a list of dataset packages,
#' `list_resources()` provides a more comprehensive and flexible
#' interface for exploring the PHS Open Data platform. It returns both
#' datasets and their associated resources in a single tibble, and
#' supports filtering by dataset titles or resource names.

#'
#' @return A tibble.
#' @export
#'
#' @examplesIf isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)
#' head(list_datasets())
#' @seealso [list_resources()]
list_datasets <- function() {
  # fetch the data
  content <- phs_GET("package_list", "")

  data_sets <- tibble::tibble(name = unlist(content$result))

  return(data_sets)
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("returns more than 0 datasets", {
  expect_gte(nrow(list_datasets()), 1L)
})

test_that("returns data in the expected format", {
  data <- list_datasets()

  expect_s3_class(data, "tbl_df")
  expect_named(data, "name")
  expect_identical(dplyr::n_distinct(data[["name"]]), nrow(data))
})

#' Generates error message from an "error" element of an `httr::content` object
#'
#' @param error The "error" element of an object produced by `httr::content`.
#' @return character string
#' @noRd
#' @keywords internal
parse_error <- function(error) {
  error_message <- error$message
  error_type <- error$`__type`

  # define error for most common cases
  error_output <- paste0(
    error_type,
    ": ",
    error_message
  )

  # special case for validation errors
  if (error_type == "Validation Error") {
    error_output <- paste0(names(error[1][1]), ": ", error[1][[1]])

    # translate message for package users
    error_output <- sub("fields", "col_select", error_output, fixed = TRUE)
    error_output <- sub("q", "row_filters", error_output, fixed = TRUE)
  }

  # special case for SQL validation errors
  if (!is.null(error$info$orig)) {
    error_output <- sub("\\^", "", error$info$orig[[1]], fixed = TRUE)
    error_output <- sub("LINE 1:", "in SQL:", error_output, fixed = TRUE)
    error_output <- sub(
      "relation",
      "resource/table",
      error_output,
      fixed = TRUE
    )
  }

  return(error_output)
}

#' Use datastore dump to download whole files
#'
#' @param res_id a resource ID
#' @return dataframe containing resource records
#' @keywords internal
#' @noRd
dump_download <- function(res_id, call = rlang::caller_env()) {
  # fetch the data
  content <- suppressMessages(
    phs_GET("dump", res_id)
  )

  # if content is a web page
  if (inherits(content, "xml_document")) {
    cli::cli_abort(
      "Can't find resource with ID {.var {res_id}} in datastore.",
      call = call
    )
  }

  # return data
  return(content[, -1L])
}

skip_if_offline(host = "www.opendata.nhs.scot")

test_that("throws error for non-existent res_ids", {
  expect_error(
    dump_download("not-real"),
    regexp = "Can't find resource with ID"
  )
})

test_that("downloads full resource", {
  data <- dump_download("a794d603-95ab-4309-8c92-b48970478c14")

  expect_identical(nrow(data), 926L)
  expect_length(data, 15L)
  expect_named(data)
  expect_s3_class(data, "tbl_df")
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_dataset.R
\name{get_dataset}
\alias{get_dataset}
\title{Get Open Data resources from a dataset}
\usage{
get_dataset(
  dataset_name,
  max_resources = NULL,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = FALSE
)
}
\arguments{
\item{dataset_name}{Name of the dataset as found on the
\href{https://www.opendata.nhs.scot}{NHS Open Data platform} (character).}

\item{max_resources}{(optional) The maximum number of resources to return
(integer). If not set, all resources are returned.}

\item{rows}{(optional) Maximum number of rows to return (integer).}

\item{row_filters}{(optional) A named list or vector specifying values of columns/fields to keep (e.g., list(Date = 20220216, Sex = "Female")).}

\item{col_select}{(optional) A character vector containing the names of desired columns/fields (e.g., c("Date", "Sex")).}

\item{include_context}{(optional) If \code{TRUE}, additional information about the resource will be added as columns to the data, including the resource ID, the resource name, the creation date, and the last modified/updated date.}
}
\value{
A \link[tibble:tibble-package]{tibble} with the data.
}
\description{
Downloads multiple resources from a dataset on the NHS Open Data
platform by dataset name, with optional row limits and context columns.
}
\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
\dontrun{
get_dataset("gp-practice-populations", max_resources = 2, rows = 10)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=get_resource]{get_resource()}} for downloading a single resource from a dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_datasets.R
\name{list_datasets}
\alias{list_datasets}
\title{Lists all available datasets}
\usage{
list_datasets()
}
\value{
A tibble.
}
\description{
\code{list_datasets()} shows all of the datasets hosted on the PHS Open Data
Platform.
}
\section{Lifecycle}{


\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{list_datasets()} has been superseded by \code{list_resources()}.
While \code{list_datasets()} only returns a list of dataset packages,
\code{list_resources()} provides a more comprehensive and flexible
interface for exploring the PHS Open Data platform. It returns both
datasets and their associated resources in a single tibble, and
supports filtering by dataset titles or resource names.
}

\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
head(list_datasets())
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=list_resources]{list_resources()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_latest_resource.R
\name{get_latest_resource}
\alias{get_latest_resource}
\title{Get the latest resource from a data set}
\usage{
get_latest_resource(
  dataset_name,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = TRUE
)
}
\arguments{
\item{dataset_name}{Name of the dataset as found on the
\href{https://www.opendata.nhs.scot}{NHS Open Data platform} (character).}

\item{rows}{(optional) Maximum number of rows to return (integer).}

\item{row_filters}{(optional) A named list or vector specifying values of columns/fields to keep (e.g., list(Date = 20220216, Sex = "Female")).}

\item{col_select}{(optional) A character vector containing the names of desired columns/fields (e.g., c("Date", "Sex")).}

\item{include_context}{(optional) If \code{TRUE}, additional information about the resource will be added as columns to the data, including the resource ID, the resource name, the creation date, and the last modified/updated date.}
}
\value{
a \link[tibble:tibble-package]{tibble} with the data
}
\description{
Returns the latest resource available in a dataset.
}
\details{
There are some datasets on the open data platform that
keep historic resources instead of updating existing ones.
For these it is useful to be able to retrieve the latest
resource. As of 1.8.2024 these data sets include:
\itemize{
\item gp-practice-populations
\item gp-practice-contact-details-and-list-sizes
\item nhsscotland-payments-to-general-practice
\item dental-practices-and-patient-registrations
\item general-practitioner-contact-details
\item prescribed-dispensed
\item dispenser-location-contact-details
\item community-pharmacy-contractor-activity
}
}
\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
\dontrun{
dataset_name <- "gp-practice-contact-details-and-list-sizes"

data <- get_latest_resource(dataset_name)

filters <- list("Postcode" = "DD11 1ES")
wanted_cols <- c("PracticeCode", "Postcode", "Dispensing")

filtered_data <- get_latest_resource(
  dataset_name = dataset_name,
  row_filters = filters,
  col_select = wanted_cols
)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_resource.R
\name{get_resource}
\alias{get_resource}
\title{Get Open Data resource}
\usage{
get_resource(
  res_id,
  rows = NULL,
  row_filters = NULL,
  col_select = NULL,
  include_context = FALSE
)
}
\arguments{
\item{res_id}{The resource ID as found on \href{https://www.opendata.nhs.scot/}{NHS Open Data platform} (character).}

\item{rows}{(optional) Maximum number of rows to return (integer).}

\item{row_filters}{(optional) A named list or vector specifying values of columns/fields to keep (e.g., list(Date = 20220216, Sex = "Female")).}

\item{col_select}{(optional) A character vector containing the names of desired columns/fields (e.g., c("Date", "Sex")).}

\item{include_context}{(optional) If \code{TRUE}, additional information about the resource will be added as columns to the data, including the resource ID, the resource name, the creation date, and the last modified/updated date.}
}
\value{
A \link[tibble:tibble-package]{tibble} with the data.
}
\description{
Downloads a single resource from the NHS Open Data platform by resource ID, with optional filtering and column selection.
}
\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
res_id <- "ca3f8e44-9a84-43d6-819c-a880b23bd278"

data <- get_resource(res_id)

filters <- list("HB" = "S08000030", "Month" = "202109")
wanted_cols <- c("HB", "Month", "TotalPatientsSeen")

filtered_data <- get_resource(
  res_id = res_id,
  row_filters = filters,
  col_select = wanted_cols
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=get_dataset]{get_dataset()}} for downloading all resources from a given dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_dataset_additional_info.R
\name{get_dataset_additional_info}
\alias{get_dataset_additional_info}
\title{get a datasets additional info}
\usage{
get_dataset_additional_info(dataset_name)
}
\arguments{
\item{dataset_name}{Name of the dataset as found on the
\href{https://www.opendata.nhs.scot}{NHS Open Data platform} (character).}
}
\value{
a \link[tibble:tibble-package]{tibble} with the data
}
\description{
\code{get_dataset_additional_info()} returns a tibble of dataset names along with
the amount of resources it has and the date it was last updated.Last updated
is taken to mean the most recent date a resource within the dataset was
created or modified.
}
\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
get_dataset_additional_info("gp-practice-populations")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_resource_sql.R
\name{get_resource_sql}
\alias{get_resource_sql}
\title{Get PHS Open Data using SQL}
\usage{
get_resource_sql(sql)
}
\arguments{
\item{sql}{A single \code{SELECT} query (character).}
}
\value{
A \link[tibble:tibble-package]{tibble} with the query results. Only 32,000 rows
can be returned from a single SQL query.
}
\description{
Downloads data from the NHS Open Data platform using a SQL query. Similar to
\code{\link[=get_resource]{get_resource()}}, but allows more flexible server-side querying. This
function has a lower maximum row number (32,000 vs 99,999) for returned
results.
}
\details{
Only 32,000 rows can be returned from a single SQL query.
}
\section{SQL syntax}{

The resource ID must be double-quoted, e.g.
\verb{SELECT * FROM "58527343-a930-4058-bf9e-3c6e5cb04010"}, as must column names,
e.g. \code{"Year"}. Strings require single quotes, e.g. \code{'value'}. This syntax is
needed because the CKAN DataStore uses
\href{https://www.postgresql.org/docs/9.2/index.html}{PostgreSQL}.
}

\section{R raw strings}{

Enclosing the query in an R raw string avoids the need to escape embedded
quotes, e.g. \verb{\\"TotalCancelled\\"}. Square brackets are the recommended
delimiter, i.e. \code{r"[...]"}, because \verb{)"} within a query, e.g.
\code{SUM("TotalCancelled")}, would prematurely close the string. Another option
is \code{r"{...}"}, in the rare case that the query contains \verb{]"}.
}

\examples{
\dontshow{if (isTRUE(length(curl::nslookup("www.opendata.nhs.scot", error = FALSE)) > 0L)) withAutoprint(\{ # examplesIf}
# Basic query
cancelled_ops <- get_resource_sql(r"[
SELECT
    "TotalCancelled",
    "TotalOperations",
    "Hospital",
    "Month"
FROM
    "bcc860a4-49f4-4232-a76b-f559cf6eb885"
WHERE
    "Hospital" = 'D102H'
]")

# Joining two resources
hb_pop <- get_resource_sql(r"[
SELECT
    pops."Year",
    pops."HB",
    lookup."HBName",
    pops."AllAges"
FROM
    "27a72cc8-d6d8-430c-8b4f-3109a9ceadb1" AS pops
    JOIN "652ff726-e676-4a20-abda-435b98dd7bdc" AS lookup ON pops."HB" = lookup."HB"
WHERE
    pops."Sex" = 'All'
    AND pops."Year" > 2006
]")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=get_resource]{get_resource()}} for downloading a resource without using a SQL query.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_resources.R
\name{list_resources}
\alias{list_resources}
\title{Provides an overview of all resources available on the Open Data platform.}
\usage{
list_resources(
  dataset_contains = NULL,
  resource_contains = NULL,
  dataset_name = lifecycle::deprecated()
)
}
\arguments{
\item{dataset_contains}{A character string containing an expression to be
used as search criteria against the dataset name.}

\item{resource_contains}{A character string containing a \href{https://www.geeksforgeeks.org/dsa/write-regular-expressions/}{regular expression}
to be matched against available resource names.}

\item{dataset_name}{Deprecated. Use dataset_contains instead.}
}
\value{
A \link[tibble:tibble-package]{tibble} containing details of all available datasets and
resources, or those containing the string specified in the
\code{dataset_contains} and \code{resource_contains} arguments.
}
\description{
Provides an overview of all resources available from
\href{https://www.opendata.nhs.scot/}{opendata.nhs.scot}, with the option to limit results based on both dataset
and resource names. The returned tibble can be used to look-up dataset
and resource ids, and is useful for exploring the available data sets.
}
\examples{

list_resources()
list_resources(dataset_contains = "standard-populations")
list_resources(
  dataset_contains = "standard-populations", resource_contains = "European"
)

}
